#pragma once
#include <petscdmplex.h>
#include <petscsnes.h>
#include <petscds.h>

#include "appctx.h"
#include "stokes.h"

#undef __FUNCT__
#define __FUNCT__ "DMVecViewLocal"
PetscErrorCode DMVecViewLocal(DM dm, Vec v, PetscViewer viewer)
{
    Vec            lv;
    PetscInt       p;
    PetscMPIInt    rank, numProcs;
    PetscErrorCode ierr;

    PetscFunctionBeginUser;
    ierr = MPI_Comm_rank(PetscObjectComm((PetscObject)dm), &rank);CHKERRQ(ierr);
    ierr = MPI_Comm_size(PetscObjectComm((PetscObject)dm), &numProcs);CHKERRQ(ierr);
    ierr = DMGetLocalVector(dm, &lv);CHKERRQ(ierr);
    ierr = DMGlobalToLocalBegin(dm, v, INSERT_VALUES, lv);CHKERRQ(ierr);
    ierr = DMGlobalToLocalEnd(dm, v, INSERT_VALUES, lv);CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD, "Local function\n");CHKERRQ(ierr);
    for (p = 0; p < numProcs; ++p) {
        if (p == rank) {ierr = VecView(lv, PETSC_VIEWER_STDOUT_SELF);CHKERRQ(ierr);}
        ierr = PetscBarrier((PetscObject) dm);CHKERRQ(ierr);
    }
    ierr = DMRestoreLocalVector(dm, &lv);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}

PetscErrorCode zero_scalar(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)
{
    u[0] = 0.0;
    return 0;
}
PetscErrorCode zero_vector(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void *ctx)
{
    PetscInt d;
    for (d = 0; d < dim; ++d) u[d] = 0.0;
    return 0;
}

PetscErrorCode constant_p(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *p, void *ctx)
{
    *p = 1.0;
    return 0;
}

#undef __FUNCT__
#define __FUNCT__ "CreatePressureNullSpace"
PetscErrorCode CreatePressureNullSpace(DM dm, AppCtx *user, Vec *v, MatNullSpace *nullSpace)
{
    Vec              vec;
    PetscErrorCode (*funcs[2])(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void* ctx) = {zero_vector, constant_p};
    PetscErrorCode   ierr;

    PetscFunctionBeginUser;
    ierr = DMGetGlobalVector(dm, &vec);CHKERRQ(ierr);
    ierr = DMProjectFunction(dm, 0.0, funcs, NULL, INSERT_ALL_VALUES, vec);CHKERRQ(ierr);
    ierr = VecNormalize(vec, NULL);CHKERRQ(ierr);
    if (user->showSolution) {
        ierr = PetscPrintf(PetscObjectComm((PetscObject)dm), "Pressure Null Space\n");CHKERRQ(ierr);
        ierr = VecView(vec, PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
    }
    ierr = MatNullSpaceCreate(PetscObjectComm((PetscObject)dm), PETSC_FALSE, 1, &vec, nullSpace);CHKERRQ(ierr);
    if (v) {
        ierr = DMCreateGlobalVector(dm, v);CHKERRQ(ierr);
        ierr = VecCopy(vec, *v);CHKERRQ(ierr);
    }
    ierr = DMRestoreGlobalVector(dm, &vec);CHKERRQ(ierr);
    /* New style for field null spaces */
    {
        PetscObject  pressure;
        MatNullSpace nullSpacePres;

        ierr = DMGetField(dm, 1, &pressure);CHKERRQ(ierr);
        ierr = MatNullSpaceCreate(PetscObjectComm(pressure), PETSC_TRUE, 0, NULL, &nullSpacePres);CHKERRQ(ierr);
        ierr = PetscObjectCompose(pressure, "nullspace", (PetscObject) nullSpacePres);CHKERRQ(ierr);
        ierr = MatNullSpaceDestroy(&nullSpacePres);CHKERRQ(ierr);
    }
    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ProcessOptions"
PetscErrorCode ProcessOptions(MPI_Comm comm, AppCtx *user)
{
    PetscErrorCode ierr;
    
    PetscFunctionBeginUser;
    user->dim             = 2;
    user->interpolate     = PETSC_FALSE;
    user->simplex         = PETSC_TRUE;
    user->refinementLimit = 0.0;
    user->testPartition   = PETSC_FALSE;
    user->showInitial     = PETSC_FALSE;
    user->showSolution    = PETSC_FALSE;
    user->showError       = PETSC_FALSE;
    user->saveVTK         = PETSC_TRUE;
    user->computeL2error  = PETSC_FALSE;
    user->setNullSpace    = PETSC_FALSE;
    user->test_ms         = PETSC_FALSE;
    
    ierr = PetscOptionsBegin(comm, "", "FFMCM Problem Options", "DMPLEX");CHKERRQ(ierr);
    ierr = PetscOptionsBool("-test_ms", "Test solver with Manufactured Solution", "ffmcm.c", user->test_ms, &user->test_ms, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsInt("-dim", "The topological mesh dimension", "ffmcm.c", user->dim, &user->dim, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-interpolate", "Generate intermediate mesh elements", "ffmcm.c", user->interpolate, &user->interpolate, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-simplex", "Use simplices or tensor product cells", "ffmcm.c", user->simplex, &user->simplex, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsReal("-refinement_limit", "The largest allowable cell volume", "ffmcm.c", user->refinementLimit, &user->refinementLimit, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-test_partition", "Use a fixed partition for testing", "ffmcm.c", user->testPartition, &user->testPartition, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-set_null_space", "Set null space of ", "ffmcm.c", user->setNullSpace, &user->setNullSpace, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-show_initial", "Output the initial guess for verification", "ffmcm.c", user->showInitial, &user->showInitial, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-show_solution", "Output the solution for verification", "ffmcm.c", user->showSolution, &user->showSolution, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-show_error", "Output the error for verification", "ffmcm.c", user->showError, &user->showError, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-compute_L2_error", "Compute the error in the L2 norm", "ffmcm.c", user->computeL2error, &user->computeL2error, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsEnd();

    ierr = PetscOptionsGetString(PETSC_NULL, PETSC_NULL, "-save_solution_vtk", user->saveSolutionVTK, PETSC_MAX_PATH_LEN, &user->saveVTK); CHKERRQ(ierr);
    ierr = PetscOptionsGetString(PETSC_NULL, PETSC_NULL, "-gmsh_file", user->gmsh_file, PETSC_MAX_PATH_LEN, NULL); CHKERRQ(ierr);
    ierr = PetscLogEventRegister("CreateMesh", DM_CLASSID, &user->createMeshEvent);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}