static char help[] = "Framework for Multiscale Computational Mechanics. \n\n";

#include "petscdmplex.h"
#include "petscsnes.h"
#include "petscds.h"

#include "stokes.h"

#include "appctx.h"
#include "meshes.h"
#include "aux.h"

#undef __FUNCT__
#define __FUNCT__ "GeneralSetup"
PetscErrorCode GeneralSetup(MPI_Comm comm, GeneralProblem *prob)
{
    PetscErrorCode ierr = 0;
    PetscBool stk=PETSC_FALSE, nav_stk=PETSC_FALSE, gmsh=PETSC_FALSE;
    char gmsh_file[PETSC_MAX_PATH_LEN];
    
    PetscFunctionBegin;
    ierr = PetscOptionsBegin(comm, "", "General Problem Options", "FMCM");CHKERRQ(ierr);
    ierr = PetscOptionsBool("-stokes", "Solves Stokes problem", "fmcm.c", stk, &stk, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsBool("-navier_stokes", "Solves Navier-Stokes problem", "fmcm.c", nav_stk, &nav_stk, NULL);CHKERRQ(ierr);
    ierr = PetscOptionsEnd();
    if (nav_stk) {
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Navier-Stokes problem under development. Try with -stokes instead. \n"); CHKERRQ(ierr);
        PetscFunctionReturn(1);
    }
    else { // Default.
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Solves Stokes problem.\n"); CHKERRQ(ierr);
        prob->SetupDiscretization = Stokes_SetupDiscretization;
        prob->SetupProblem = Stokes_Setup;
    }
    
    ierr = PetscOptionsGetString(PETSC_NULL, PETSC_NULL, "-gmsh_file", gmsh_file, PETSC_MAX_PATH_LEN, &gmsh); CHKERRQ(ierr);
    if(gmsh) {
        prob->CreateMesh = CreateMesh_Gmsh;
        PetscFunctionReturn(0);
    }
    else { // default
        prob->CreateMesh = CreateMesh_PETSc;
        PetscFunctionReturn(0);
    }
    
    PetscFunctionReturn(1);
}

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc, char **argv)
{
    SNES           snes;                 /* nonlinear solver */
    DM             dm;                   /* problem definition */
    Vec            u,r;                  /* solution, residual vectors */
    Mat            A,J;                  /* Jacobian matrix */
    MatNullSpace   nullSpace;            /* May be necessary for pressure */
    AppCtx         user;                 /* user-defined work context */
    PetscInt       its;                  /* iterations for convergence */
    PetscReal      error         = 0.0;  /* L_2 error in the solution */
    PetscReal      ferrors[2];
    PetscErrorCode ierr;
    GeneralProblem prob;

    ierr = PetscInitialize(&argc, &argv, NULL, help);CHKERRQ(ierr);
    ierr = GeneralSetup(PETSC_COMM_WORLD, &prob); CHKERRQ(ierr);
    ierr = ProcessOptions(PETSC_COMM_WORLD, &user);CHKERRQ(ierr);
    ierr = SNESCreate(PETSC_COMM_WORLD, &snes);CHKERRQ(ierr);
    ierr = SetupMesh(PETSC_COMM_WORLD,&user, prob.CreateMesh, &dm);CHKERRQ(ierr);
    ierr = SNESSetDM(snes, dm);CHKERRQ(ierr);
    ierr = DMSetApplicationContext(dm, &user);CHKERRQ(ierr);

    ierr = prob.SetupDiscretization(dm, &user);CHKERRQ(ierr);
    ierr = DMPlexCreateClosureIndex(dm, NULL);CHKERRQ(ierr);
    
    ierr = DMCreateGlobalVector(dm, &u);CHKERRQ(ierr);
    ierr = VecDuplicate(u, &r);CHKERRQ(ierr);
    
    ierr = DMSetMatType(dm,MATAIJ);CHKERRQ(ierr);
    ierr = DMCreateMatrix(dm, &J);CHKERRQ(ierr);
    A = J;
    if(user.setNullSpace) {
        ierr = CreatePressureNullSpace(dm, &user, NULL, &nullSpace);CHKERRQ(ierr);
        ierr = MatSetNullSpace(A, nullSpace);CHKERRQ(ierr);
    }

    ierr = DMPlexSetSNESLocalFEM(dm,&user,&user,&user);CHKERRQ(ierr);
    ierr = SNESSetJacobian(snes, A, J, NULL, NULL);CHKERRQ(ierr);
    
    ierr = SNESSetFromOptions(snes);CHKERRQ(ierr);
    
    if (user.showInitial) {ierr = DMVecViewLocal(dm, u, PETSC_VIEWER_STDOUT_SELF);CHKERRQ(ierr);}

    if (user.showSolution) {
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Initial guess\n");CHKERRQ(ierr);
        ierr = VecView(u, PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
    }
    if(user.test_ms) {
        PetscReal res;
        ierr = DMProjectFunction(dm, 0.0, user.exactFuncs, NULL, INSERT_ALL_VALUES, u);CHKERRQ(ierr);
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Exact Solution\n"); CHKERRQ(ierr);
        ierr = VecView(u, PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
        ierr = SNESComputeFunction(snes, u, r); CHKERRQ(ierr);
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Initial Residual\n"); CHKERRQ(ierr);
        ierr = VecView(r, PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
        ierr = VecNorm(r, NORM_2, &res); CHKERRQ(ierr);
        ierr = PetscPrintf(PETSC_COMM_WORLD, "L_2 Residual: %g\n", res); CHKERRQ(ierr);
        {
            Vec          b;
            PetscBool    isNull;
            
            ierr = SNESComputeJacobian(snes, u, A, A); CHKERRQ(ierr);
            ierr = MatNullSpaceTest(nullSpace, J, &isNull); CHKERRQ(ierr);
            if (!isNull) SETERRQ(PETSC_COMM_WORLD, PETSC_ERR_PLIB, "The null space calculated for the system operator is invalid.");
            ierr = VecDuplicate(u, &b); CHKERRQ(ierr);
            ierr = VecSet(r, 0.0); CHKERRQ(ierr);
            ierr = SNESComputeFunction(snes, r, b); CHKERRQ(ierr);
            ierr = MatMult(A, u, r); CHKERRQ(ierr);
            ierr = VecAXPY(r, 1.0, b); CHKERRQ(ierr);
            ierr = VecDestroy(&b); CHKERRQ(ierr);
            ierr = PetscPrintf(PETSC_COMM_WORLD, "Au - b = Au + F(0)\n"); CHKERRQ(ierr);
            ierr = VecView(r, PETSC_VIEWER_STDOUT_WORLD); CHKERRQ(ierr);
            ierr = VecNorm(r, NORM_2, &res); CHKERRQ(ierr);
            ierr = PetscPrintf(PETSC_COMM_WORLD, "Linear L_2 Residual: %g\n", res); CHKERRQ(ierr);
        }
    }

    PetscErrorCode (*initialGuess[2])(PetscInt dim, PetscReal time, const PetscReal x[], PetscInt Nf, PetscScalar *u, void* ctx) = {zero_vector, zero_scalar};
    ierr = DMProjectFunction(dm, 0.0, initialGuess, NULL, INSERT_VALUES, u);CHKERRQ(ierr);
    ierr = SNESSolve(snes, NULL, u);CHKERRQ(ierr);
    
    ierr = SNESGetIterationNumber(snes, &its);CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD, "Number of SNES iterations = %D\n", its);CHKERRQ(ierr);
    if(user.computeL2error) {
        ierr = DMComputeL2Diff(dm, 0.0, user.exactFuncs, NULL, u, &error);CHKERRQ(ierr);
        ierr = DMComputeL2FieldDiff(dm, 0.0, user.exactFuncs, NULL, u, ferrors);CHKERRQ(ierr);
        ierr = PetscPrintf(PETSC_COMM_WORLD, "L_2 Error: %.3g [%.3g, %.3g]\n", error, ferrors[0], ferrors[1]);CHKERRQ(ierr);
    }
    if (user.showError) {
        Vec r;
        ierr = DMGetGlobalVector(dm, &r);CHKERRQ(ierr);
        ierr = DMProjectFunction(dm, 0.0, user.exactFuncs, NULL, INSERT_ALL_VALUES, r);CHKERRQ(ierr);
        ierr = VecAXPY(r, -1.0, u);CHKERRQ(ierr);
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Solution Error\n");CHKERRQ(ierr);
        ierr = VecView(r, PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
        ierr = DMRestoreGlobalVector(dm, &r);CHKERRQ(ierr);
    }
    if (user.showSolution) {
        ierr = PetscPrintf(PETSC_COMM_WORLD, "Solution\n");CHKERRQ(ierr);
        ierr = VecChop(u, 3.0e-9);CHKERRQ(ierr);
        ierr = VecView(u, PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);
    }

    if (user.saveVTK) {
        PetscViewer viewer;
        ierr = PetscViewerVTKOpen(PETSC_COMM_WORLD, user.saveSolutionVTK, FILE_MODE_WRITE, &viewer); CHKERRQ(ierr);
        ierr = VecView(u,viewer); CHKERRQ(ierr);
        ierr = PetscViewerDestroy(&viewer); CHKERRQ(ierr);
    }
    
    if(nullSpace) ierr = MatNullSpaceDestroy(&nullSpace);CHKERRQ(ierr);
    if (A != J) {ierr = MatDestroy(&A);CHKERRQ(ierr);}
    ierr = MatDestroy(&J);CHKERRQ(ierr);
    ierr = VecDestroy(&u);CHKERRQ(ierr);
    ierr = VecDestroy(&r);CHKERRQ(ierr);
    ierr = SNESDestroy(&snes);CHKERRQ(ierr);
    ierr = DMDestroy(&dm);CHKERRQ(ierr);
    ierr = PetscFinalize();
    return 0;
}
