#pragma once
#include <petscdmplex.h>
#include <petscsnes.h>
#include <petscds.h>

#include "appctx.h"

#undef __FUNCT__
#define __FUNCT__ "CheckMeshTopology"
static PetscErrorCode CheckMeshTopology(DM dm)
{
    PetscInt       dim, coneSize, cStart;
    PetscBool      isSimplex;
    PetscErrorCode ierr;
    
    PetscFunctionBeginUser;
    ierr = DMGetDimension(dm, &dim);CHKERRQ(ierr);
    ierr = DMPlexGetHeightStratum(dm, 0, &cStart, NULL);CHKERRQ(ierr);
    ierr = DMPlexGetConeSize(dm, cStart, &coneSize);CHKERRQ(ierr);
    isSimplex = coneSize == dim+1 ? PETSC_TRUE : PETSC_FALSE;
    ierr = DMPlexCheckSymmetry(dm);CHKERRQ(ierr);
    ierr = DMPlexCheckSkeleton(dm, isSimplex, 0);CHKERRQ(ierr);
    ierr = DMPlexCheckFaces(dm, isSimplex, 0);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CheckGeometryAndReverseCells"
PetscErrorCode CheckGeometryAndReverseCells(DM dm)
{
    PetscInt       dim, coneSize, cStart, cEnd, c, inv_elem=0;
    PetscReal     *v0, *J, *invJ, detJ;
    PetscErrorCode ierr;
    
    PetscFunctionBeginUser;
    ierr = DMGetDimension(dm, &dim);CHKERRQ(ierr);
    ierr = DMPlexGetHeightStratum(dm, 0, &cStart, &cEnd);CHKERRQ(ierr);
    ierr = DMPlexGetConeSize(dm, cStart, &coneSize);CHKERRQ(ierr);
    ierr = PetscMalloc3(dim,&v0,dim*dim,&J,dim*dim,&invJ);CHKERRQ(ierr);
    for (c = cStart; c < cEnd; ++c) {
        ierr = DMPlexComputeCellGeometryFEM(dm, c, NULL, v0, J, invJ, &detJ);CHKERRQ(ierr);
        //if (detJ <= 0.0) SETERRQ2(PETSC_COMM_SELF, PETSC_ERR_ARG_OUTOFRANGE, "Invalid determinant %g for cell %d", detJ, c);
        if (detJ <= 0.0) { ierr = DMPlexReverseCell(dm, c); CHKERRQ(ierr); inv_elem++; }
    }
    ierr = PetscFree3(v0,J,invJ);CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD, "CheckGeometryAndReverseCells() : Number of inverted cells %d\n", inv_elem);CHKERRQ(ierr);
    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CreateMesh_PETSc"
PetscErrorCode CreateMesh_PETSc(MPI_Comm comm, AppCtx *user, DM *dm)
{
    PetscInt       dim             = user->dim;
    PetscBool      interpolate     = user->interpolate;
    PetscReal      refinementLimit = user->refinementLimit;
    const PetscInt cells[3]        = {3, 3, 3};
    PetscErrorCode ierr;
    
    PetscFunctionBeginUser;
    if (user->simplex) {ierr = DMPlexCreateBoxMesh(comm, dim, interpolate, dm);CHKERRQ(ierr);}
    else               {ierr = DMPlexCreateHexBoxMesh(comm, dim, cells, DM_BOUNDARY_NONE, DM_BOUNDARY_NONE, DM_BOUNDARY_NONE, dm);CHKERRQ(ierr);}
    {
        DM refinedMesh     = NULL;

        /* Refine mesh using a volume constraint */
        ierr = DMPlexSetRefinementLimit(*dm, refinementLimit);CHKERRQ(ierr);
        if (user->simplex) {ierr = DMRefine(*dm, comm, &refinedMesh);CHKERRQ(ierr);}
        if (refinedMesh) {
            ierr = DMDestroy(dm);CHKERRQ(ierr);
            *dm  = refinedMesh;
        }
        /* Setup test partitioning */
        if (user->testPartition) {
            PetscInt         triSizes_n2[2]       = {4, 4};
            PetscInt         triPoints_n2[8]      = {3, 5, 6, 7, 0, 1, 2, 4};
            PetscInt         triSizes_n3[3]       = {2, 3, 3};
            PetscInt         triPoints_n3[8]      = {3, 5, 1, 6, 7, 0, 2, 4};
            PetscInt         triSizes_n5[5]       = {1, 2, 2, 1, 2};
            PetscInt         triPoints_n5[8]      = {3, 5, 6, 4, 7, 0, 1, 2};
            PetscInt         triSizes_ref_n2[2]   = {8, 8};
            PetscInt         triPoints_ref_n2[16] = {1, 5, 6, 7, 10, 11, 14, 15, 0, 2, 3, 4, 8, 9, 12, 13};
            PetscInt         triSizes_ref_n3[3]   = {5, 6, 5};
            PetscInt         triPoints_ref_n3[16] = {1, 7, 10, 14, 15, 2, 6, 8, 11, 12, 13, 0, 3, 4, 5, 9};
            PetscInt         triSizes_ref_n5[5]   = {3, 4, 3, 3, 3};
            PetscInt         triPoints_ref_n5[16] = {1, 7, 10, 2, 11, 13, 14, 5, 6, 15, 0, 8, 9, 3, 4, 12};
            const PetscInt  *sizes = NULL;
            const PetscInt  *points = NULL;
            PetscPartitioner part;
            PetscInt         cEnd;
            PetscMPIInt      rank, numProcs;
            
            ierr = MPI_Comm_rank(comm, &rank);CHKERRQ(ierr);
            ierr = MPI_Comm_size(comm, &numProcs);CHKERRQ(ierr);
            ierr = DMPlexGetHeightStratum(*dm, 0, NULL, &cEnd);CHKERRQ(ierr);
            if (!rank) {
                if (dim == 2 && user->simplex && numProcs == 2 && cEnd == 8) {
                    sizes = triSizes_n2; points = triPoints_n2;
                } else if (dim == 2 && user->simplex && numProcs == 3 && cEnd == 8) {
                    sizes = triSizes_n3; points = triPoints_n3;
                } else if (dim == 2 && user->simplex && numProcs == 5 && cEnd == 8) {
                    sizes = triSizes_n5; points = triPoints_n5;
                } else if (dim == 2 && user->simplex && numProcs == 2 && cEnd == 16) {
                    sizes = triSizes_ref_n2; points = triPoints_ref_n2;
                } else if (dim == 2 && user->simplex && numProcs == 3 && cEnd == 16) {
                    sizes = triSizes_ref_n3; points = triPoints_ref_n3;
                } else if (dim == 2 && user->simplex && numProcs == 5 && cEnd == 16) {
                    sizes = triSizes_ref_n5; points = triPoints_ref_n5;
                } else SETERRQ(comm, PETSC_ERR_ARG_WRONG, "No stored partition matching run parameters");
            }
            ierr = DMPlexGetPartitioner(*dm, &part);CHKERRQ(ierr);
            ierr = PetscPartitionerSetType(part, PETSCPARTITIONERSHELL);CHKERRQ(ierr);
            ierr = PetscPartitionerShellSetPartition(part, numProcs, sizes, points);CHKERRQ(ierr);
        }
    }
    ierr = PetscStrncpy(user->bndLabelName, "marker", 7); CHKERRQ(ierr);
    user->bndLabelName[6] = 0;
    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CreateMesh_Gmsh"
PetscErrorCode CreateMesh_Gmsh(MPI_Comm comm, AppCtx *user, DM *dm)
{
    PetscErrorCode ierr=0;
    PetscFunctionBeginUser;
    ierr = DMPlexCreateGmshFromFile(comm, user->gmsh_file, user->interpolate, dm); CHKERRQ(ierr);
    ierr = PetscStrncpy(user->bndLabelName, "Face Sets", 10); CHKERRQ(ierr);
    user->bndLabelName[9] = 0;
    PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SetupMesh"
PetscErrorCode SetupMesh(MPI_Comm comm, AppCtx *user, PetscErrorCode(*create_mesh)(MPI_Comm, AppCtx*, DM*), DM *dm)
{
    PetscErrorCode ierr=0;
    PetscFunctionBeginUser;
    ierr = PetscLogEventBegin(user->createMeshEvent,0,0,0,0);CHKERRQ(ierr);
    ierr = create_mesh(comm, user, dm); CHKERRQ(ierr);
    {
        DM distributedMesh = NULL;
        ierr = DMPlexDistribute(*dm, 0, NULL, &distributedMesh);CHKERRQ(ierr);
        if (distributedMesh) {
            ierr = DMDestroy(dm);CHKERRQ(ierr);
            *dm  = distributedMesh;
        }
    }
    ierr = DMSetFromOptions(*dm);CHKERRQ(ierr);
    ierr = DMViewFromOptions(*dm, NULL, "-dm_view");CHKERRQ(ierr);
    ierr = CheckMeshTopology(*dm);CHKERRQ(ierr);
    ierr = CheckGeometryAndReverseCells(*dm); CHKERRQ(ierr);
    ierr = PetscLogEventEnd(user->createMeshEvent,0,0,0,0);CHKERRQ(ierr);
    
    /* Homogeneus dirichlet boundary */
    PetscInt      dim = user->dim;
    PetscInt      hom_dirichlet_id[10];
    /* Non-homogeneous dirichlet boundary */
    PetscInt      nonhom_dirichlet_id[10];
   /* Neumann boundaries */
    PetscInt      neumann_id[10];

    /* 
     ------------------------------
     Homogeneous Dirichlet boundary
     ------------------------------
    */
    
    // [FT] question: should user->hom_dirichlet_id be preallocated?
    ierr = PetscOptionsGetIntArray(PETSC_NULL, PETSC_NULL, "-hom_dirichlet_id", hom_dirichlet_id, &user->num_hom_dirichlet_id, PETSC_NULL); CHKERRQ(ierr);
    if(user->num_hom_dirichlet_id != 0) { ierr = PetscMalloc(user->num_hom_dirichlet_id*sizeof(PetscInt), &user->hom_dirichlet_id); CHKERRQ(ierr); }
    for(PetscInt i=0;i<user->num_hom_dirichlet_id;i++) user->hom_dirichlet_id[i]=hom_dirichlet_id[i];
    
    /*
     ----------------------------------
     Non-homogeneous Dirichlet boundary
     ----------------------------------
     */
    
    ierr = PetscOptionsGetIntArray(PETSC_NULL, PETSC_NULL, "-nonhom_dirichlet_id", nonhom_dirichlet_id, &user->num_nonhom_dirichlet_id, PETSC_NULL); CHKERRQ(ierr);
    if(user->num_nonhom_dirichlet_id != 0) { ierr = PetscMalloc(user->num_nonhom_dirichlet_id*sizeof(PetscInt), &user->nonhom_dirichlet_id); CHKERRQ(ierr); }
    for(PetscInt i=0;i<user->num_nonhom_dirichlet_id;i++) user->nonhom_dirichlet_id[i]=nonhom_dirichlet_id[i];

    ierr = PetscMalloc(dim*sizeof(PetscInt), &user->nonhom_dirichlet_values); CHKERRQ(ierr);
    ierr = PetscOptionsGetRealArray(PETSC_NULL, PETSC_NULL, "-nonhom_dirichlet_values", user->nonhom_dirichlet_values, &dim, PETSC_NULL); CHKERRQ(ierr);
    
    /*
     ---------------
     Neuman boundary
     ---------------
    */
    
    ierr = PetscOptionsGetIntArray(PETSC_NULL, PETSC_NULL, "-neumann_id", neumann_id, &user->num_neumann_id, PETSC_NULL); CHKERRQ(ierr);
    if(user->num_neumann_id != 0) {
        ierr = PetscMalloc(user->num_neumann_id*sizeof(PetscInt), &user->neumann_id); CHKERRQ(ierr);
        ierr = PetscMalloc(user->num_neumann_id*sizeof(PetscReal), &user->neumann_values); CHKERRQ(ierr);
        
        ierr = PetscOptionsGetRealArray(PETSC_NULL, PETSC_NULL, "-neumann_values", user->neumann_values, &user->num_neumann_id, PETSC_NULL); CHKERRQ(ierr);
    }
    for(PetscInt i=0;i<user->num_neumann_id;i++) user->neumann_id[i]=neumann_id[i];

    /*
     --------------------
     Checking consistency
     --------------------
     */
    if(user->num_hom_dirichlet_id == 0 &&
       user->num_neumann_id == 0) {
        ierr = PetscPrintf(PETSC_COMM_WORLD, "FFMCM::Stokes_ProcessOptions()[WARNING] No Dirichlet boundary condition was defined.\n");CHKERRQ(ierr);
        ierr = PetscPrintf(PETSC_COMM_WORLD, "FFMCM::Stokes_ProcessOptions()[WARNING] You may need to define -set_null_space.\n");CHKERRQ(ierr);
    }

    PetscFunctionReturn(0);
}
